import logging
from typing import Dict


logger = logging.getLogger("controller")


def watch_for_mutations(mutations: dict, namespace: str):
    from kubernetes import client, config, watch  # type: ignore
    from pprint import pformat

    config.load_incluster_config()
    while True:
        try:
            api = client.CustomObjectsApi()
            w = watch.Watch()
            for event in w.stream(
                api.list_namespaced_custom_object,
                group="apps.dominodatalab.com",
                version="v1alpha1",
                plural="mutations",
                namespace=namespace,
            ):
                try:
                    logger.debug(pformat(event))
                    if event["type"] == "ADDED":
                        name = event["object"]["metadata"]["name"]
                        logger.warning(f"Mutation added: {name}")
                        mutations[name] = event["object"]["rules"]
                    elif event["type"] == "MODIFIED":
                        logger.info(f"Mutation modified: {name}")
                        mutations[name] = event["object"]["rules"]
                    elif event["type"] == "DELETED":
                        name = event["object"]["metadata"]["name"]
                        logger.warning(f"Mutation deleted: {name}")
                        del mutations[name]
                    else:
                        logger.warn(
                            f'Unrecognized event type: {event["type"]}'
                        )
                except Exception as e:
                    logger.exception(
                        f"Exception while handling: {e}\n{pformat(event)}"
                    )
            logger.warning("Watcher exited without error; restarting.")
        except Exception as e:
            logger.error(f"Watcher exited with error {e}; restarting.")


def start(namespace: str) -> dict:
    import threading

    logger.info("Starting Informer Thread.")
    mutations: Dict[str, dict] = {}
    thread = threading.Thread(
        target=watch_for_mutations, args=(mutations, namespace)
    )
    thread.start()
    return mutations


if __name__ == "__main__":
    import time

    logging.basicConfig(
        level=logging.DEBUG,
        format="%(asctime)s - %(name)s - %(levelname)s - %(message)s",
    )
    mutations = start("domino-platform")
    while True:
        time.sleep(5)
        print(mutations)
