from typing import Dict, Any
import copy
import json

PROXY_CONTAINER_NAME = "istio-proxy"
PROXY_STATUS_PORT = 15020
KUBE_APP_PROBER_ENV_NAME = "ISTIO_KUBE_APP_PROBERS"


def extract_port_map(container) -> Dict[str, int]:
    port_map = {}
    for port in container.get("ports", []):
        if "name" in port and port["name"] != "":
            port_map[port["name"]] = port["containerPort"]
    return port_map


def update_named_port(probe, port_map):
    if isinstance(probe["port"], str):
        if port := port_map[probe["port"]]:
            probe["port"] = port


def patch_rewrite_probe(
    probe_config,
    container,
    port_map: Dict[str, int],
    key: str,
    path: str,
    status_port: int,
):
    if probe := container.get(key, {}).get("httpGet"):
        update_named_port(probe, port_map)
        probe_config[path] = copy.deepcopy(container[key])
        probe["port"] = status_port
        probe["path"] = path
        probe["scheme"] = "HTTP"
    else:
        return


def find_sidecar(spec):
    for i, container in enumerate(spec.get("containers", [])):
        if container["name"] == PROXY_CONTAINER_NAME:
            return i
    return None


def istio_probe_rewrite(modified_spec: dict, config: dict):
    meta = modified_spec["request"]["object"]["metadata"]
    spec = modified_spec["request"]["object"]["spec"]
    meta["annotations"] = meta.get("annotations", {})

    rewrite_app_http_probers = meta["annotations"].get(
        "sidecar.istio.io/rewriteAppHTTPProbers", "true"
    )

    if rewrite_app_http_probers == "false":
        return

    if (sidecar_idx := find_sidecar(spec)) is None:
        return

    spec["containers"][sidecar_idx]["env"] = spec["containers"][
        sidecar_idx
    ].get("env", [])

    for var in spec["containers"][sidecar_idx]["env"]:
        if var["name"] == KUBE_APP_PROBER_ENV_NAME:
            return

    status_port = int(
        meta["annotations"].get(
            "status.sidecar.istio.io/port", PROXY_STATUS_PORT
        )
    )

    probe_config: Dict[str, Any] = {}

    for container in spec["containers"]:
        name = container["name"]

        if name == PROXY_CONTAINER_NAME:
            continue

        port_map = extract_port_map(container)

        patch_rewrite_probe(
            probe_config,
            container,
            port_map,
            "livenessProbe",
            f"/app-health/{name}/livez",
            status_port,
        )
        patch_rewrite_probe(
            probe_config,
            container,
            port_map,
            "readinessProbe",
            f"/app-health/{name}/readyz",
            status_port,
        )
        patch_rewrite_probe(
            probe_config,
            container,
            port_map,
            "startupProbe",
            f"/app-health/{name}/startupz",
            status_port,
        )

    spec["containers"][sidecar_idx]["env"].append(
        {"name": KUBE_APP_PROBER_ENV_NAME, "value": json.dumps(probe_config)}
    )
