from mutation import WORKLOAD_CONTAINERS
from typing import List, Optional
import logging

logger = logging.getLogger(__name__)


def istio_port_exception(modified_spec: dict, config: dict):
    def add_annotation(metadata: dict, key: str, ports: List[int]):
        combined_ports = [port for port in ports]

        if key in metadata["annotations"]:
            combined_ports.extend(
                map(
                    lambda port: int(port),
                    metadata["annotations"][key].split(","),
                )
            )

        combined_ports_s = map(lambda port: str(port), set(combined_ports))
        value = ",".join(combined_ports_s)
        metadata["annotations"][key] = value

    def add_args(container: dict, key: str, ports: List[int]):
        combined_ports = [port for port in ports]
        index = None
        for i, arg in enumerate(container["args"]):
            if arg == key:
                index = i
                break

        if index is not None:
            combined_ports.extend(
                map(
                    lambda port: int(port),
                    container["args"][index + 1].split(","),
                )
            )

        combined_ports_s = map(lambda port: str(port), set(combined_ports))

        value = ",".join(combined_ports_s)

        if index is not None:
            container["args"][index + 1] = value
        else:
            container["args"].extend([key, value])

    metadata = modified_spec["request"]["object"]["metadata"]
    metadata["annotations"] = metadata.get("annotations", {})
    spec = modified_spec["request"]["object"]["spec"]

    inbound_ports: Optional[List[int]] = config.get("inbound")
    outbound_ports: Optional[List[int]] = config.get("outbound")

    if config.get("spark", False):
        from kubernetes import client, config  # type: ignore
        from kubernetes.client.rest import ApiException  # type: ignore

        config.load_incluster_config()  # type: ignore
        kubeclient = client.CoreV1Api()
        namespace = modified_spec["request"]["namespace"]
        execution_id = metadata.get("labels", {}).get(
            "dominodatalab.com/execution-id"
        )
        if execution_id is not None:
            nodeport_service_name = f"run-{execution_id}-nodeport"
            try:
                # Fetch nodeport
                service = kubeclient.read_namespaced_service(
                    nodeport_service_name, namespace
                )
                if inbound_ports is None:
                    inbound_ports = []
                if outbound_ports is None:
                    outbound_ports = []
                # Add to ports
                for port in service.spec.ports:
                    inbound_ports.append(port.node_port)
                    outbound_ports.append(port.node_port)
            except ApiException as e:
                logger.warning(f"Unable to read nodeport service: {e})")
        else:
            logger.warning("Unable to determine execution id.")

    inbound_key = "traffic.sidecar.istio.io/excludeInboundPorts"
    if inbound_ports is not None and len(inbound_ports) > 0:
        add_annotation(metadata, inbound_key, inbound_ports)

        for container in spec.get("initContainers", []):
            name = container["name"]
            if name == "istio-init":
                add_args(container, "-d", inbound_ports)

    outbound_key = "traffic.sidecar.istio.io/excludeOutboundPorts"
    if outbound_ports is not None and len(outbound_ports) > 0:
        add_annotation(metadata, outbound_key, outbound_ports)

        for container in spec.get("initContainers", []):
            name = container["name"]
            if name == "istio-init":
                add_args(container, "-o", outbound_ports)


def service_account(modified_spec: dict, config: dict):
    workload_metadata = modified_spec["request"]["object"]["metadata"]
    username = workload_metadata["labels"].get(
        "dominodatalab.com/starting-user-username"
    )
    if username is None:
        logger.warning("No user found.")
        return
    logger.debug(f"Found user {username}.")
    execution_id = workload_metadata["labels"].get(
        "dominodatalab.com/execution-id"
    )
    if execution_id is None:
        logger.warning("No execution_id found.")
        return
    from kubernetes.client import V1ServiceAccount  # type: ignore
    from kubernetes import client, config  # type: ignore
    from kubernetes.client.rest import ApiException  # type: ignore
    import time

    config.load_incluster_config()  # type: ignore
    kubeclient = client.CoreV1Api()

    namespace = modified_spec["request"]["namespace"]

    try:
        result = kubeclient.read_namespaced_service_account(
            username, namespace
        )
        secrets = result.secrets
        logger.debug("Service account already exists.")
    except ApiException as e:
        if e.status == 404:
            logger.debug("Service account not found; creating.")
            # Create SA
            body = V1ServiceAccount(
                metadata={"name": username, "namespace": namespace}
            )
            result = kubeclient.create_namespaced_service_account(
                namespace, body
            )
            logger.debug(f"Created service account: {result}.")
            # Wait for secret to be generated.
            while True:
                result = kubeclient.read_namespaced_service_account(
                    username, namespace
                )
                secrets = result.secrets
                if secrets is not None and len(secrets) > 0:
                    break
                time.sleep(0.5)
        else:
            logger.exception(e)
            raise

    secret_name = secrets[0].name

    # Create role binding
    binding_name = f"run-{execution_id}"
    kubeclient = client.RbacAuthorizationV1Api()
    body = [
        {
            "op": "add",
            "path": "/subjects/-",
            "value": {
                "kind": "ServiceAccount",
                "name": username,
                "namespace": namespace,
            },
        }
    ]
    result = kubeclient.patch_namespaced_role_binding(
        binding_name, namespace, body
    )
    logger.debug(f"Created role binding: {result}.")

    # Update Pod Spec with SA
    spec = modified_spec["request"]["object"]["spec"]
    spec["serviceAccountName"] = username

    # Update secret volume
    volumes = spec["volumes"]
    for volume in volumes:
        if volume["name"] == "service-account-token-vol":
            volume["secret"]["secretName"] = secret_name


def autoscaling(modified_spec: dict, config: dict):
    spec = modified_spec["request"]["object"]["spec"]
    spec["worker"]["autoscaling"] = spec["worker"].get("autoscaling", {})
    autoscaling = spec["worker"]["autoscaling"]
    autoscaling["enabled"] = True
    autoscaling["replicasMax"] = config["replicasMax"]
    autoscaling["CpuTargetPercentage"] = config["CpuTargetPercentage"]


def shared_memory(modified_spec: dict, config: dict):
    spec = modified_spec["request"]["object"]["spec"]
    if "volumes" not in spec:
        spec["volumes"] = []

    volumes = spec["volumes"]
    should_add = True
    for volume in volumes:
        if volume["name"] == "dshm":
            should_add = False

    if should_add:
        spec["volumes"].append(
            {
                "name": "dshm",
                "emptyDir": {"medium": "Memory", "sizeLimit": config["limit"]},
            }
        )

    for container in spec["containers"]:
        name = container["name"]
        if name in WORKLOAD_CONTAINERS:
            if "volumeMounts" not in container:
                container["volumeMounts"] = []
            container["volumeMounts"].append(
                {"name": "dshm", "mountPath": "/dev/shm"}
            )


def gpu_toleration(modified_spec: dict, config: dict):
    spec = modified_spec["request"]["object"]["spec"]
    for toleration in spec["tolerations"]:
        if toleration["key"] == "nvidia.com/gpu":
            return

    spec["tolerations"].append(
        {"key": "nvidia.com/gpu", "operator": "Exists", "effect": "NoExecute"}
    )

    request = config["request"]

    for container in spec["containers"]:
        if "spark" in container["name"]:
            container["resources"]["limits"]["nvidia.com/gpu"] = request
            container["resources"]["requests"]["nvidia.com/gpu"] = request


def pvc(modified_spec: dict, config: dict):
    spec = modified_spec["request"]["object"]["spec"]

    # Upsert Volumes
    if "volumes" not in spec:
        spec["volumes"] = []

    volumes = spec["volumes"]
    should_add = True
    for volume in volumes:
        if volume["name"] == config["claimName"]:
            should_add = False
    if should_add:
        volumes.append(
            {
                "name": config["claimName"],
                "persistentVolumeClaim": {"claimName": config["claimName"]},
            }
        )

    for container in modified_spec["request"]["object"]["spec"]["containers"]:
        name = container["name"]
        if name in WORKLOAD_CONTAINERS:
            if "volumeMounts" not in container:
                container["volumeMounts"] = []

            should_add = True
            for mount in container["volumeMounts"]:
                if mount["name"] == config["claimName"]:
                    should_add = False

            if should_add:
                container["volumeMounts"].append(
                    {"name": config["claimName"], "mountPath": config["path"]}
                )


def fuse(modified_spec: dict, config: dict):
    spec = modified_spec["request"]["object"]["spec"]
    metadata = modified_spec["request"]["object"]["metadata"]

    # Upsert Volumes
    if "volumes" not in spec:
        spec["volumes"] = []

    volumes = spec["volumes"]
    should_add = True
    for volume in volumes:
        if volume["name"] == "fuse":
            should_add = False
    if should_add:
        volumes.append(
            {
                "name": "fuse",
                "hostPath": {
                    "path": "/dev/fuse",
                },
            }
        )

    # Upsert Annotations
    metadata["annotations"] = metadata.get("annotations", {})

    for container in spec["containers"]:
        name = container["name"]
        if name in WORKLOAD_CONTAINERS:
            if "volumeMounts" not in container:
                container["volumeMounts"] = []

            should_add = True
            for mount in container["volumeMounts"]:
                if mount["name"] == "fuse":
                    should_add = False

            if should_add:
                container["volumeMounts"].append(
                    {"name": "fuse", "mountPath": "/dev/fuse"}
                )

            # Upsert Security Context
            container["securityContext"] = container.get("securityContext", {})
            container["securityContext"]["privileged"] = True
            container["securityContext"]["capabilities"] = container[
                "securityContext"
            ].get("capabilities", {})
            capabilities = container["securityContext"]["capabilities"]
            capabilities["add"] = list(
                set(
                    capabilities.get("add", [])
                    + ["SYS_ADMIN", "DAC_READ_SEARCH", "MKNOD"]
                )
            )

            annotations = metadata["annotations"]
            key = f"container.apparmor.security.beta.kubernetes.io/{name}"
            annotations[key] = "unconfined"
