#!/usr/bin/env python
import logging
from pymongo import MongoClient  # type: ignore
from flask import Flask, request, jsonify  # type: ignore
from pprint import pformat
import base64
import copy
import jsonpatch  # type: ignore
from typing import List, Callable, Dict
import os
import re
from urllib.parse import quote_plus
from mutations import (
    fuse,
    pvc,
    gpu_toleration,
    shared_memory,
    autoscaling,
    istio_port_exception,
    service_account,
)
from mutation.efa import efa
from mutation.container_image_rewrite import container_image_rewrite
from mutation.resources import resources
from mutation.insert_container import insert_container
from mutation.insert_volumes import insert_volumes
from mutation.insert_volume_mounts import insert_volume_mounts
from mutation.insert_execution_jwt import insert_execution_jwt
from mutation.modify_annotation import modify_annotation
from mutation.modify_label import modify_label
from mutation.istio_probe_rewrite import istio_probe_rewrite
from mutation.toleration import add_toleration
from mutation.modify_env import modify_env


logger = logging.getLogger("webhook")
app = Flask(__name__)

SUPPORTED_MUTATIONS = {
    "Pod": {
        "execution": {
            "fuse": fuse,
            "gpuToleration": gpu_toleration,
            "pvc": pvc,
            "sharedMemory": shared_memory,
            "istioPortException": istio_port_exception,
            "efa": efa,
            "containerImageRewrite": container_image_rewrite,
            "resources": resources,
            "serviceAccount": service_account,
            "insertContainer": insert_container,
            "insertVolumes": insert_volumes,
            "insertVolumeMounts": insert_volume_mounts,
            "insertExecutionJWT": insert_execution_jwt,
            "modifyAnnotation": modify_annotation,
            "modifyLabel": modify_label,
            "istioProbeRewrite": istio_probe_rewrite,
            "addToleration": add_toleration,
            "modifyEnv": modify_env,
        },
        "build": {
            "resources": resources,
            "insertContainer": insert_container,
            "insertVolumes": insert_volumes,
            "insertVolumeMounts": insert_volume_mounts,
            "modifyAnnotation": modify_annotation,
            "modifyLabel": modify_label,
        },
    },
    "SparkCluster": {
        "autoscaling": autoscaling,
        "modifyAnnotation": modify_annotation,
        "modifyLabel": modify_label,
    },
}

DEFAULT_PLATFORM_NAMESPACE = "domino-platform"


equality_selector = re.compile(r"^([^\s=!]+)\s*([=!]{1,2})\s*([^\s=!]+)$")
set_selector = re.compile(r"^(\S+)\s*(notin|in)\s*\((\S+)\)$")


def parse_selector(s: str) -> Callable[[Dict[str, str]], bool]:
    # Set Selector
    if match := set_selector.match(s):
        key = match.group(1)
        operator = match.group(2)
        values = set(map(lambda s: s.strip(), match.group(3).split(",")))

        def set_func(labels: Dict[str, str]) -> bool:
            value = labels.get(key)
            result = value in values
            if operator == "notin":
                result = not result
            return result

        return set_func

    # Equality Selector
    if match := equality_selector.match(s):
        key = match.group(1)
        operator = match.group(2)
        value = match.group(3)
        if operator not in ["==", "!=", "="]:
            raise ValueError("Equality operator {operator} not ==, !=, or =.")

        def equality_func(labels: Dict[str, str]) -> bool:
            label_value = labels.get(key)
            result = label_value == value
            if operator == "!=":
                result = not result
            return result

        return equality_func

    # Key Exists
    def func(labels: Dict[str, str]) -> bool:
        return s in labels

    return func


def fetch_orgs(user_id: str) -> List[str]:
    if os.environ.get("MONGO_PASSWORD") is None:
        return []
    logger.debug(f"Finding organizations for user '{user_id}'.")
    if user_id is None:
        return []
    platform_namespace = os.environ.get(
        "PLATFORM_NAMESPACE", DEFAULT_PLATFORM_NAMESPACE
    )
    host = os.environ.get(
        "MONGO_HOST",
        f"mongodb-replicaset.{platform_namespace}.svc.cluster.local:27017",
    )
    username = quote_plus(os.environ.get("MONGO_USERNAME", "admin"))
    password = quote_plus(os.environ["MONGO_PASSWORD"])
    db_name = quote_plus(os.environ.get("MONGO_DB_NAME", "domino"))
    if username == "admin":
        path = ""
    else:
        path = "/{}".format(db_name)
    mongo_uri = "mongodb://{}:{}@{}{}".format(username, password, host, path)
    with MongoClient(mongo_uri) as client:
        user = client[db_name]["users"].find_one({"loginId.id": user_id})
        _id = user["_id"]
        logger.debug(f"Found user id {_id}")
        pipeline = [
            {"$match": {"members": {"$elemMatch": {"userId": _id}}}},
            {
                "$lookup": {
                    "from": "users",
                    "localField": "organizationUserId",
                    "foreignField": "_id",
                    "as": "users",
                }
            },
        ]
        orgs = list(client[db_name]["organizations"].aggregate(pipeline))
        orgs = [org["users"][0]["loginId"]["id"] for org in orgs]
    logger.debug(f"Found organizations: {orgs}")
    return orgs


@app.route("/healthz")
def alive():
    return "{'status': 'Healthy'}"


def match_namespace(namespace: str, rule: dict) -> bool:
    return (
        "namespaceSelector" not in rule
        or namespace in rule["namespaceSelector"]
    )


def match_project(project: str, rule: dict) -> bool:
    return "projectSelector" not in rule or project in rule["projectSelector"]


def match_tier(tier: str, rule: dict) -> bool:
    return (
        "hardwareTierIdSelector" not in rule
        or tier in rule["hardwareTierIdSelector"]
    )


def match_username(username: str, rule: dict) -> bool:
    return (
        "usernameSelector" not in rule or username in rule["usernameSelector"]
    )


def match_orgs(orgs: List[str], rule: dict) -> bool:
    return (
        "organizationSelector" not in rule
        or len(set(orgs).intersection(set(rule["organizationSelector"]))) > 0
    )


def match_build(is_build: bool, rule: dict) -> bool:
    build_selector = rule.get("matchBuilds", False)
    return is_build == build_selector


def match_labels(labels: Dict[str, str], rule: dict) -> bool:
    labelSelectors = rule.get("labelSelectors", [])
    selectors = list(map(parse_selector, labelSelectors))
    return all(map(lambda selector: selector(labels), selectors))


def match_self(modified_spec: dict) -> bool:
    # Returns false if object is DomSed pod.
    spec = modified_spec["request"]["object"]["spec"]
    containers = spec.get("containers", [])
    for container in containers:
        if container.get("name") == "domsed":
            return False
    return True


@app.route("/", methods=["POST"])
def webhook():
    request_info = request.json
    modified_spec = copy.deepcopy(request_info)
    workload_metadata = modified_spec["request"]["object"]["metadata"]
    labels = workload_metadata.get("labels", {})
    workload_type = modified_spec["request"]["kind"]["kind"]
    name = workload_metadata.get("generateName")
    if name is None:
        name = workload_metadata["name"]
    namespace = modified_spec["request"]["namespace"]

    is_build = (
        workload_metadata.get("annotations", {}).get(
            "imagebuilder.dominodatalab.com/version"
        )
        == "v2"
    )

    tier = workload_metadata["labels"].get(
        "dominodatalab.com/hardware-tier-id"
    )
    project = workload_metadata["labels"].get("dominodatalab.com/project-name")
    project_owner = workload_metadata["labels"].get(
        "dominodatalab.com/project-owner-username"
    )
    project = f"{project_owner}/{project}"
    username = workload_metadata["labels"].get(
        "dominodatalab.com/starting-user-username"
    )
    orgs = fetch_orgs(username)

    logger.warning(
        (
            f"Processing {workload_type} {namespace}/{name}."
            f"\tTier: {tier}\tProject: {project}\tProject Owner: "
            f"{project_owner}\tUsername: {username}\tOrgs: {orgs}"
            f"\tBuild: {is_build}"
        )
    )

    for name, rules in MUTATIONS.items():
        pre_modified_spec = copy.deepcopy(modified_spec)
        try:
            for i, rule in enumerate(rules):
                if (
                    match_project(project, rule)
                    and match_tier(tier, rule)
                    and match_namespace(namespace, rule)
                    and match_username(username, rule)
                    and match_orgs(orgs, rule)
                    and match_build(is_build, rule)
                    and match_labels(labels, rule)
                    and match_self(modified_spec)
                ):
                    logger.warning(f"Mutation {name} rule {i} matched.")
                    logger.debug(pformat(rule))

                    supported = SUPPORTED_MUTATIONS[workload_type]
                    if workload_type == "Pod":
                        if is_build:
                            supported = supported["build"]
                        else:
                            supported = supported["execution"]

                    for k in supported:
                        if k in rule:
                            logger.info(
                                f"Applying {k} mutation: {pformat(rule[k])}"
                            )
                            func = supported[k]
                            func(modified_spec, copy.deepcopy(rule[k]))
                        else:
                            logger.debug(f"Rule does not have mutation {k}")
                else:
                    logger.info(f"Mutation {name} rule {i} did not match.")

        except Exception as e:
            # Rollback
            logger.exception(e)
            logger.warning(f"Mutation {name} failed, rolling back.")
            modified_spec = pre_modified_spec

    patch = jsonpatch.JsonPatch.from_diff(
        request_info["request"]["object"], modified_spec["request"]["object"]
    )

    logger.warning(str(patch))

    admission_response = {
        "allowed": True,
        "uid": request_info["request"]["uid"],
        "patch": base64.b64encode(str(patch).encode()).decode(),
        "patchtype": "JSONPatch",
    }
    admissionReview = {"response": admission_response}

    logger.debug(pformat(admissionReview))
    return jsonify(admissionReview)


if __name__ == "__main__":
    from controller import start

    lvl = logging.getLevelName(os.environ.get("LOG_LEVEL", "WARNING"))
    logging.basicConfig(
        level=lvl,
        format="%(asctime)s - %(name)s - %(levelname)s - %(message)s",
    )
    log = logging.getLogger("werkzeug")
    log.setLevel(logging.WARNING)

    if os.environ.get("MONGO_PASSWORD") is None:
        logger.warning("No MONGO_PASSWORD, organization filter disabled.")

    platform_namespace = os.environ.get(
        "PLATFORM_NAMESPACE", DEFAULT_PLATFORM_NAMESPACE
    )

    MUTATIONS = start(platform_namespace)
    debug = os.environ.get("FLASK_ENV") == "development"
    app.run(
        host=os.environ.get("FLASK_HOST", "0.0.0.0"),
        port=5000,
        debug=debug,
        ssl_context=("/ssl/cert.pem", "/ssl/key.pem"),
    )
